      subroutine cmpt_ctf(fft_klh, hkl_max, fft_ann,
     $     t_analy, mem_cmpt, ctf_val, abs_ctf, disef,
     $     fno, weight, nptles, indx, theta, phi, omega,
     $     x, y, mag, score, orient_flag, origin_flag,
     $     delf_maj_ptle, delf_min_ptle, ang_maj_ptle)

C     This routine searches 'defocus space' to determine the major
C     defocus, minor defocus,and astigmatism orientation that results in
C     the best overall agreement between the particle images in a
C     micrograph and the model. For the special case of stigmated
C     images, where the major and minor defoci are equal, the search is
C     done over a single defocus parameter and the orientation is kept
C     at zero.
C
C     Although the image/model comparison can be done using any choice
C     of funcmode, the phase method (funcmode=3) works best since it
C     just uses the phase information to determine the optimal location
C     of the CTF nodes.

      implicit none

#ifdef MPI
      include 'mpif.h'
#endif

      include '../include/sizes.inc'
      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'

      include '../include/pifhead.inc'

      integer, intent(in) :: hkl_max, fft_ann, fno, nptles
      real, intent(in) :: weight
      complex, intent(in) :: fft_klh(-fft_ann:fft_ann, -fft_ann:fft_ann, 
     $     0:fft_ann)
      real, intent(out) :: ctf_val(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(out) :: abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(in)  :: disef(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)

      integer, intent(in) :: indx(nptles)
      real, intent(in)  :: theta(nptles), phi(nptles), omega(nptles)
      real, intent(in)  :: x(nptles), y(nptles), mag(nptles)
      real, intent(out) :: score(nptles)
      character, intent(out) :: orient_flag(nptles), origin_flag(nptles)
      real, intent(out) :: delf_maj_ptle(nptles), delf_min_ptle(nptles)
      real, intent(out) :: ang_maj_ptle(nptles)

      double precision, intent(out) :: mem_cmpt
      double precision, intent(inout) :: t_analy

      integer err_code
      integer defmaj_step, defmaj_nangle
      integer defmin_step, defmin_nangle
      integer angle_step,  angle_nangle

      integer i, j, k, max_chunk, indx1, indx2
      integer mem_size, start_view, end_view,
     +     tmp_row, radius, nsize
      integer min_ctfs(3)
      integer, dimension(3):: steps
      integer alloc_stat, tmp_nangle

      real :: imgmagfactor
      real :: r4_max = huge(r4_max)
      real du, dv
      real center(2), num_annul, value, min_annul, ctfs(3)
      double precision part_dist
      real    rot_mat(3,3), 
     +     dist(-Nfocus:Nfocus,-Nfocus:Nfocus,-Nangle:Nangle),
     +     sum_dist(-Nfocus:Nfocus,-Nfocus:Nfocus,-Nangle:Nangle)

      real    pixels(npix_row, npix_col)
      
      integer, allocatable :: invalid(:)
      real, allocatable :: rffta_omega(:)
      complex, allocatable :: fc_2d_ref(:,:),
     +  fc_2d_intp(:,:,:), fc_2d(:,:,:)
      
      double precision elapsed_time
      double precision time

      integer clock1, clock2
      
      double precision :: mem_orient = 0.0D0,
     +     mem_rffta_omega = 0.0D0, mem_fc_2d = 0.0D0 

      character(len=12) :: subname = 'cmpt_ctfr'

#ifdef MPI
      integer stat(mpi_status_size)
      integer buffer_size
#endif

C     Initializations
      steps(1) = Nfocus
      steps(2) = Nfocus
      steps(3) = Nangle
      
C     Set loop bounds for anastigmatic and astigmatic cases
      if(anastigmatic == 1) then
         defmaj_nangle = Nfocus
         defmin_nangle = 0
         angle_nangle  = 0
      else
         defmaj_nangle = Nfocus
         defmin_nangle = Nfocus
         angle_nangle  = nangle
      endif

C     Allocate storage for subroutine FFTANALY
      allocate(rffta_omega(nrffta_omega), stat = alloc_stat)
      
      mem_rffta_omega = dble(4 * nrffta_omega)
      mem_orient      = dble(4*10*max_views)
      
      nsize    = (2*Nfocus+1)*(2*Nfocus+1)*(2*Nangle+1)
      mem_cmpt = mem_orient + mem_rffta_omega + nsize

      if(alloc_stat > 0) then
         call error_stop (progname, subname, 
     +        ' Memory allocation failed for FFT arrays')
      endif

C     to initialize the value of rffta_omega(nrfft_col+15) that is
C     rffta_omega(nrffta_omega) Initialize the FFT parameters used to
C     transform rows of RFFTA each row has length NRFFTA_COL
      call vrffti( nrffta_col, rffta_omega )

      call system_clock(clock1)

      if (Nangle == 0 .and. Nfocus == 0) goto 1111
      
      num_annul = real(nrffta_row)*angstrom_per_pix/res_max 
      min_annul = real(nrffta_row)*angstrom_per_pix/res_min
      if (num_annul .ge. nrffta_row/2.) 
     +     call error_stop(progname,subname, 
     +     'Resolution is too high, smaller than 2*angstrom_per_pix')
C      if (iam == 0) then
C         print *,min_annul,num_annul,Dfocus,Nfocus,Dangle,Nangle
C      endif
      radius = int(num_annul)+1

C     In the original version of this routine, the minimum radius used
C     for comparing the images and projections was recomputed on the
C     basis of the defocus value. This can result in the optimal defocus
C     depending on the starting defocus. Although the effect is small,
C     this has been noticed in some cases where the defocus step size is
C     also small. The following block of code has been commented out and
C     retained in case the feature is ever reimplemented.

ccc      real    wavelength, tmaj, fw2
ccc      wavelength = 1.2264264967
ccc     +     /sqrt(VOLTS(fno)+ 0.9784663543E-6 * VOLTS(fno)**2)
ccc      tmaj = min(DELF_MAJS(fno),DELF_MINS(fno))
ccc      fw2 = (tmaj-sqrt(tmaj*tmaj - AMPFACS(fno)*wavelength*2.0))
ccc     +     /AMPFACS(fno)
ccc      min_annul=sqrt(fw2/1000.)*(PIXSIZS(fno)*2*hkl_max)/wavelength/10.
ccc      if (min_annul < 10.) min_annul= 10.
ccc      if (iam == 0) then
ccc         print *, 'change the minimum radius to be', min_annul
ccc      endif

C     read each views and generate 2DDFT for these views
      call set_indices(num_nodes, 1, nptles,
     +         frst_lst_view, max_chunk )

C     allocate memory to store cuts,  DFT bakup,
      mem_size        = ( 1 + nfc_2d_row ) * ( 1 + 2 * nfc_2d_col )
      allocate(fc_2d_ref(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col), 
     +  fc_2d_intp(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col,0:max_chunk-1),
     +  fc_2d(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col,0:max_chunk-1),
     +  invalid(0:max_chunk-1),
     +  stat = alloc_stat)
      
      if(alloc_stat > 0) then
         call error_stop (progname, subname, 
     +        ' Memory allocation failed for fc_2d arrays')
      endif

      mem_fc_2d       = dble( (16 * max_chunk + 8) * mem_size)
      mem_cmpt        = mem_cmpt + mem_fc_2d

C     Initialize output parameters
      score(1:nptles) = r4_max
      orient_flag(1:nptles) = 'S'
      origin_flag(1:nptles) = 'S'
      delf_maj_ptle(1:nptles) = 0.
      delf_min_ptle(1:nptles) = 0.
      ang_maj_ptle(1:nptles)  = 0.
      
C     Read particle images and calculate transforms
      do j = 0, max_chunk - 1

         do i = num_nodes - 1 , 0 , -1
            start_view = frst_lst_view(1,i) + j
            end_view   = frst_lst_view(2,i)
            if (start_view <= end_view ) then

               if(iam == i) then

                  call gen_read_pixel_frame(pix_file_name, 
     $                 pix_file_unit,
     $                 indx(start_view), npix_col, npix_row, 
     $                 npix_col, npix_row, 1, pixels, 
     $                 err_code)

                  if(err_code > 0) then
                     invalid(j) = 1
                     cycle
                  else
                     invalid(j) = 0
                  endif

               endif

            endif
         enddo

         start_view = frst_lst_view(1,iam) + j
         end_view   = frst_lst_view(2,iam)
         if (start_view > end_view) exit !Break out of loop

         center(1)    = x(start_view)
         center(2)    = y(start_view)
C     discard particles not well aligned
         if(  center(1) < npix_row*0.1 .or. 
     $     center(1) > npix_row*0.9 .or.
     $     center(2) < npix_col*0.1 .or. 
     $     center(2) > npix_col*0.9) then
             invalid(j) = 1
             orient_flag(start_view) = 'N'
             origin_flag(start_view) = 'N'
             score(start_view) = 999.99
             cycle
         endif

C     adjust the imgmagfactor according to its own pixel_size
         imgmagfactor = mag(start_view) * PIXSIZ_OUT /
     $        angstrom_per_pix

C     Compute Discrete Fourier Transform of pixel values:

         call fft_2dfft(pixels, npix_row, npix_col, center,
     +        nrffta_omega, rffta_omega, nrffta_row, 
     +        nrffta_col, nfc_2d_row, 
     +        nfc_2d_col, fc_2d(:,:,j) )

         du = center(1) - real(int(center(1)))
         dv = center(2) - real(int(center(2)))
         call shift_fft(fc_2d(:,:,j), 
     +        nfc_2d_row, nfc_2d_col, du, dv)


C     Compute cut for this view:

         call crowther_to_matrix(theta(start_view), -phi(start_view), 
     $        -omega(start_view), rot_mat)
         call interpl_3d(fft_klh, hkl_max, min_annul, num_annul, rot_mat
     $        , fc_2d_intp(:,:,j), imgmagfactor, fft_ann)
      enddo

      call system_clock(clock2)
      time = elapsed_time(clock1, clock2)
      t_analy = t_analy + time

C     Initialize entire array of scores to large value
      dist(:,:,:) = R4_MAX

C     loop to refine CTF, I use min_ctfs and ctfs the temp value of CTF
C     the resolution for defocus is Dfocus, and the resolution for angle
C     is Dangle

 1110 min_ctfs(1) = 0           ! defocus major
      min_ctfs(2) = 0           ! defocus minor
      min_ctfs(3) = 0           ! the angle
 
      do defmaj_step = -defmaj_nangle, defmaj_nangle  
         ctfs(1)= Dfocus*defmaj_step + DELF_MAJS(fno)
         if (ctfs(1) <= 0) cycle ! Skip to next iteration
         do defmin_step = -defmin_nangle, defmin_nangle  

            if(anastigmatic==1) then
               ctfs(2) = ctfs(1)
            else
               ctfs(2)= Dfocus*defmin_step + DELF_MINS(fno)
               if (ctfs(2) <= 0) cycle ! Skip to next iteration
            endif

C     If ctfs(1) equals ctfs(2), then don't need to rotate about
C     angle. To avoid having to reproduce a big chunk of code, set the
C     loop indices based on the result of this test.
            if (ctfs(2) == ctfs(1)) then
               tmp_nangle = 0
            else
               tmp_nangle = angle_nangle
            endif

            do angle_step = -tmp_nangle, tmp_nangle
               ctfs(3) = dangle*angle_step + ANG_MAJS(fno)
               if (dist(defmaj_step,defmin_step,angle_step) < 
     +              R4_MAX-1) cycle
               IF (CTFMODE > 0 .and. CTFMODE <= 4)  then
                  call ctf_para(ctf_val, abs_ctf, 
     +                 0, nfc_2d_row, -nfc_2d_col, nfc_2d_col, .true.,
     +                 CTFMODE, Filter, TEMPFAC, CTF_FF1, CTF_FF2,
     +                 PIXSIZS(fno), units, VOLTS(fno),ctfs(1), 
     +                 ctfs(2), ctfs(3), AMPFACS(fno),
     +                 CS_COEFFS(fno), Weight, res_max,res_max)
               endif

               part_dist = 0.0
               start_view = frst_lst_view(1,iam)
               end_view   = frst_lst_view(2,iam)

               do j = start_view, end_view
                  k = j - start_view
                  if(invalid(k) == 1) cycle
                  fc_2d_ref = fc_2d_intp(:,:,k) * ctf_val

                  call imgcomport(fc_2d_ref, hkl_max, 
     +                 fc_2d(:,:,k), disef, abs_ctf, 
     +                 min_annul, num_annul, value, funcmode)
                  part_dist = part_dist + value
                  if (value < score(j)) then
                     score(j) = value
                     delf_maj_ptle(j) = ctfs(1)
                     delf_min_ptle(j) = ctfs(2)
                     ang_maj_ptle(j)  = ctfs(3)
                  endif
               enddo

               dist(defmaj_step,defmin_step,angle_step)=real(part_dist)

            enddo
         enddo
      enddo
      
#ifdef MPI
      call MPI_Allreduce(dist, sum_dist, nsize,
     +     MPI_REAL, MPI_SUM, MPI_COMM_WORLD, mpi_err)
#else
      sum_dist = dist           ! Fortran 90 array syntax
#endif
      
C     Determine location of minimum element in array sum_dist. Note use
C     of Fortran 90 array syntax and intrinsic function minloc (which
C     behaves as if lower bounds of array are all equal to 1)

      min_ctfs = minloc(sum_dist)
      min_ctfs = min_ctfs - steps - 1

      if ( abs(real(min_ctfs(1))) > Nfocus*0.5 .or. 
     +     abs(real(min_ctfs(2))) > Nfocus*0.5 .or.
     +     abs(real(min_ctfs(3))) > Nangle*0.5 ) then
         if(anastigmatic==1) then
            DELF_MAJS(fno) = DELF_MAJS(fno) + min_ctfs(1) * Dfocus
            DELF_MINS(fno) = DELF_MAJS(fno)
            ANG_MAJS(fno)  = 0.0
         else
            DELF_MAJS(fno) = DELF_MAJS(fno) + min_ctfs(1) * Dfocus
            DELF_MINS(fno) = DELF_MINS(fno) + min_ctfs(2) * Dfocus
            ANG_MAJS(fno)  = ANG_MAJS(fno)  + min_ctfs(3) * Dangle
         endif
         
C     slide dist windows to make min dist value is in center
         call recenter_array_3d(dist, Nfocus, Nfocus, Nangle,
     +                          min_ctfs, R4_MAX)
         
         goto  1110
      endif

      if(anastigmatic==1) then
         DELF_MAJS(fno) = DELF_MAJS(fno) + min_ctfs(1) * Dfocus
         DELF_MINS(fno) = DELF_MAJS(fno)
         ANG_MAJS(fno)  = 0.0
      else
         DELF_MAJS(fno) = DELF_MAJS(fno) + min_ctfs(1) * Dfocus
         DELF_MINS(fno) = DELF_MINS(fno) + min_ctfs(2) * Dfocus
         ANG_MAJS(fno)  = ANG_MAJS(fno)  + min_ctfs(3) * Dangle
      endif
      
C     Collect results back onto root process
#ifdef MPI
      do i = 1, num_nodes - 1
         start_view = frst_lst_view(1,i)
         end_view   = frst_lst_view(2,i)
         
         if (start_view > end_view) cycle
         buffer_size = end_view - start_view + 1

         if(iam == i) then
            call mpi_send(score(start_view), buffer_size,
     $         mpi_real, 0, i, MPI_COMM_WORLD, mpi_err)
            call mpi_send(delf_maj_ptle(start_view), buffer_size,
     $         mpi_real, 0, i, MPI_COMM_WORLD, mpi_err)
            call mpi_send(delf_min_ptle(start_view), buffer_size,
     $         mpi_real, 0, i, MPI_COMM_WORLD, mpi_err)
            call mpi_send(ang_maj_ptle(start_view), buffer_size,
     $         mpi_real, 0, i, MPI_COMM_WORLD, mpi_err)
            call mpi_send(origin_flag(start_view), buffer_size,
     $         mpi_character, 0, i, MPI_COMM_WORLD, mpi_err)
            call mpi_send(orient_flag(start_view), buffer_size,
     $         mpi_character, 0, i, MPI_COMM_WORLD, mpi_err)
         endif
         if(iam == 0) then
            call mpi_recv(score(start_view), buffer_size,
     $         mpi_real, i, i, MPI_COMM_WORLD, stat, mpi_err)
            call mpi_recv(delf_maj_ptle(start_view), buffer_size,
     $         mpi_real, i, i, MPI_COMM_WORLD, stat, mpi_err)
            call mpi_recv(delf_min_ptle(start_view), buffer_size,
     $         mpi_real, i, i, MPI_COMM_WORLD, stat, mpi_err)
            call mpi_recv(ang_maj_ptle(start_view), buffer_size,
     $         mpi_real, i, i, MPI_COMM_WORLD, stat, mpi_err)
            call mpi_recv(origin_flag(start_view), buffer_size,
     $         mpi_character, i, i, MPI_COMM_WORLD, stat, mpi_err)
            call mpi_recv(orient_flag(start_view), buffer_size,
     $         mpi_character, i, i, MPI_COMM_WORLD, stat, mpi_err)
         endif
      enddo
#endif
 
 1111 deallocate(rffta_omega, fc_2d, fc_2d_intp, fc_2d_ref, invalid)

      return
      end subroutine cmpt_ctf
